import numpy as np

from nav2d.assets.elements import Point, Polygon, Box
from nav2d.assets.regions import NoEntryRegion, BlackHoleRegion, SimpleRewardRegion
from nav2d.envs.base import Navigation


class HomoNav(Navigation):
    def __init__(
            self,
            init_x_normal: tuple = None,
            init_y_normal: tuple = None,
            # screen_scale: int = 600,
            # margin: float = 0.1,
            # dynamic_zones: List[DynamicRegion] = [],
            # reward_zones: List[RewardRegion] = [],
            draw_init: bool = True,
            is_eval: bool = False,
    ):
        reward_zone = SimpleRewardRegion(
            zone=Polygon([Point(3, 7.5), Point(4, 8.5), Point(8.5, 4), Point(7.5, 3)]),
            reward=-100,
        )

        init_x_normal_default = (4.5, 0.5) if is_eval else (1.0, 0.5) 
        init_y_normal_default = (4.5, 0.5) if is_eval else (7.5, 1.25) 
        init_x_normal = init_x_normal_default if init_x_normal is None else init_x_normal
        init_y_normal = init_y_normal_default if init_y_normal is None else init_y_normal
                
        super().__init__(
            init_x_normal=init_x_normal,
            init_y_normal=init_y_normal,
            reward_zones=[reward_zone],
            draw_init=draw_init,
        )

        self.goal = np.array([9., 9.])


class HeteroNav(Navigation):
    def __init__(
            self,
            init_x_normal: tuple = None,
            init_y_normal: tuple = None,
            # screen_scale: int = 600,
            # margin: float = 0.1,
            # dynamic_zones: List[DynamicRegion] = [],
            # reward_zones: List[RewardRegion] = [],
            is_eval: bool = False,
            draw_init: bool = True,
    ):
        dynamic_zone = BlackHoleRegion(
            zone=Polygon([Point(3, 5), Point(3, 8), Point(8, 8), Point(8, 3), Point(5, 3)]),
            # center=Point(5.5, 5.5),
            center=Point(4, 4),
            force=0.5,
        )
        reward_zone = SimpleRewardRegion(
            zone=Box(Point(4, 4), 3, 3),
            reward=-10,
        )

        init_x_normal_default = (4.0, 0.5) if is_eval else (1.0, 0.5) 
        init_y_normal_default = (4.0, 0.5) if is_eval else (7.5, 1.25) 
        init_x_normal = init_x_normal_default if init_x_normal is None else init_x_normal
        init_y_normal = init_y_normal_default if init_y_normal is None else init_y_normal
        
        super().__init__(
            init_x_normal=init_x_normal,
            init_y_normal=init_y_normal,
            dynamic_zones=[dynamic_zone],
            reward_zones=[reward_zone],
            draw_init=draw_init,
        )

        self.goal = np.array([9., 9.])


class WallNav(Navigation):
    def __init__(
            self,
            init_x_normal: tuple = None,
            init_y_normal: tuple = None,
            version: int = 0,       # 0 - No example zones, 1 - Three example zones
            # screen_scale: int = 600,
            # margin: float = 0.1,
            # dynamic_zones: List[DynamicRegion] = [],
            # reward_zones: List[RewardRegion] = [],
            is_eval: bool = False,
            draw_init: bool = True,
    ):
        assert version in (0, 1)
        if version == 0:
            example_zones = []
        elif version == 1:
            example_zones = [
                NoEntryRegion(
                    zone=Polygon([Point(0.5, 0.5), Point(0.8, 0.5), Point(0.8, 0.8), Point(0.5, 0.8)])
                ),  
                NoEntryRegion(
                    zone=Polygon([Point(1.2, 1.0), Point(1.4, 0.8), Point(1.6, 1.0), Point(1.4, 1.2)])
                ),  
                NoEntryRegion(
                    zone=Polygon([Point(0.8, 2.0), Point(0.7, 2.4), Point(0.9, 2.5), Point(1.0, 2.1)])
                ),
            ]
        dynamic_zone = NoEntryRegion(
            zone=Polygon([Point(3, 7.5), Point(4, 8.5), Point(8.5, 4), Point(7.5, 3)])
        )

        init_x_normal_default = (4.5, 0.5) if is_eval else (1.0, 0.5) 
        init_y_normal_default = (4.5, 0.5) if is_eval else (7.5, 1.25) 
        init_x_normal = init_x_normal_default if init_x_normal is None else init_x_normal
        init_y_normal = init_y_normal_default if init_y_normal is None else init_y_normal
        
        super().__init__(
            init_x_normal=init_x_normal,
            init_y_normal=init_y_normal,
            dynamic_zones=[dynamic_zone] + example_zones,
            draw_init=draw_init,
        )

        self.goal = np.array([9., 9.])




from abc import abstractmethod
from typing import Tuple, List, Callable
import numpy as np
import pygame

from .elements import Element, Point, Vector, Line, Polygon, RelativePos


class Region(Element):
    def __init__(self, zone: Polygon) -> None:
        self._zone = zone

    def check_segment_cross(
        self, p0: Point, proposition: Vector
    ) -> Tuple[List[Point], List[RelativePos]]:
        def __handle_corners(edge: Line, edge_next: Line):
            if edge._b == p0:
                return True
            ba = edge._a - edge._b
            bc = edge_next._b - edge_next._a
            if proposition.cross(ba) * proposition.cross(bc) > 0:
                return False
            return True

        anchor_pts: List[Point] = [p0, p0 + proposition]
        pp = Line(p0, p0 + proposition)
        for edge, edge_next in zip(
            self._zone._edges, self._zone._edges[1:] + [self._zone._edges[0]]
        ):
            cross = pp.cross_point(edge)
            if cross is True:
                overlap = pp.get_overlap(edge)
                anchor_pts.extend([overlap._a, overlap._b])

            elif isinstance(cross, Point):
                # Vertex crosspoint will only be handled on the corner.
                if cross == edge._b:
                    if any((
                        cross == p0,
                        __handle_corners(edge, edge_next),
                    )):
                        anchor_pts.append(edge._b)
                elif cross != edge._a:
                    anchor_pts.append(cross)

        anchor_pts = list(set(anchor_pts))
        lengths = [(cross - p0).length for cross in anchor_pts]
        _, anchor_pts = zip(*sorted(zip(lengths, anchor_pts), key=lambda x: x[0]))
        seg2poly_relation = []
        for p1, p2 in zip(anchor_pts[:-1], anchor_pts[1:]):
            mid = p1 + (p2 - p1) / 2
            seg2poly_relation.append(self._zone.point_relative_pos(mid))

        return anchor_pts, seg2poly_relation

    @property
    @abstractmethod
    def should_apply(self):
        ...

    def first_contact(
        self, p0: Point, propositions: List[Vector]
    ) -> Tuple[Point, float]:
        if isinstance(propositions, Vector):
            propositions = [propositions]

        if self._zone.point_relative_pos(p0) == RelativePos.IN:
            return p0, -1

        p = p0
        cuml_dist = 0
        for prop in propositions:
            anchor_pts, rel_poss = self.check_segment_cross(p, prop)
            for anchor1, anchor2, rel_pos in zip(
                anchor_pts[:-1], anchor_pts[1:], rel_poss
            ):
                if rel_pos in self.should_apply:
                    return anchor1, cuml_dist
                cuml_dist += (anchor2 - anchor1).length
            p += prop

        return None, np.inf

    def render(self, viewer, to_pixel: Callable, color: Tuple[int, int, int] = None):
        pygame.draw.polygon(
            viewer,
            self.COLOR if color is None else color,
            [to_pixel(v.pos) for v in self._zone._vertices],
            width=0,  # 0 for fill, > 0 for line thickness
        )


class DynamicRegion(Region):
    COLOR = [160, 160, 160]

    @abstractmethod
    def apply_dynamic(self, p0: Point, propositions: List[Vector]):
        ...

    @property
    def should_apply(self):
        return [RelativePos.IN]


class SimpleDynamicRegion(DynamicRegion):
    def apply_dynamic(self, p0: Point, proposition: Vector):
        if not isinstance(proposition, Vector):
            raise TypeError("SimpleDynamicRegion does not support complex movements.")
        return self._apply_dynamic(p0, proposition)

    @abstractmethod
    def _apply_dynamic(self, p0: Point, proposition: Vector):
        ...


class PunchRegion(SimpleDynamicRegion):
    """
    An extra, constant motion will be applied as long as the agent ENTERS the region.
    """

    def __init__(self, zone: Polygon, force: Vector) -> None:
        super().__init__(zone)
        self._force = force

    def _apply_dynamic(self, p0: Point, proposition: Vector):
        _, rel_poss = self.check_segment_cross(p0, proposition)
        if RelativePos.IN in rel_poss:
            return proposition + self._force
        return proposition

    def __repr__(self) -> str:
        return f"Punch zone with a fixed force '{self._force._pos}'. " + str(self._zone)


class NoEntryRegion(DynamicRegion):
    """The agent CAN walk alongside the walls."""

    COLOR = [0, 0, 0]

    def apply_dynamic(self, p0: Point, propositions: List[Vector]):
        p = p0
        movements = []
        for prop in propositions:
            anchor_pts, rel_poss = self.check_segment_cross(p, prop)
            for anchor, rel_pos in zip(anchor_pts[:-1], rel_poss):
                if rel_pos in self.should_apply:
                    movements.append(anchor - p)
                    return movements
            movements.append(prop)
            p += prop
        return movements

    def __repr__(self) -> str:
        return "No entry zone. " + str(self._zone)


class SlipperyRegion(DynamicRegion):
    def __init__(self, zone: Polygon, force: Vector) -> None:
        super().__init__(zone)
        self._force = force

    def apply_dynamic(self, p0: Point, propositions: List[Vector]):
        p = p0
        movements = []
        total_length = sum([prop.length for prop in propositions])
        for prop in propositions:
            normal_length = 0
            anchor_pts, rel_poss = self.check_segment_cross(p, prop)
            movement = None
            for anchor1, anchor2, rel_pos in zip(
                anchor_pts[:-1], anchor_pts[1:], rel_poss
            ):
                if rel_pos not in self.should_apply:
                    normal_length += (anchor2 - anchor1).length
                else:
                    delta_t = (prop.length - normal_length) / total_length
                    movement = [] if anchor1 == p else [anchor1 - p]
                    movement.append(p + prop - anchor1 + self._force * delta_t)
                    break
            movement = [prop] if movement is None else movement
            movements.extend(movement)
            for m in movement:
                p += m

        return movements

    def __repr__(self) -> str:
        return f"Slippery zone with a force rate '{self._force._pos}'. " + str(
            self._zone
        )


class BlackHoleRegion(DynamicRegion):
    def __init__(self, zone: Polygon, center: Point, force: float) -> None:
        super().__init__(zone)
        self._center = center
        self._force = force

    def apply_dynamic(self, p0: Point, propositions: List[Vector]):
        p = p0
        movements = []
        total_length = sum([prop.length for prop in propositions])
        for prop in propositions:
            normal_length = 0
            anchor_pts, rel_poss = self.check_segment_cross(p, prop)
            movement = None
            for anchor1, anchor2, rel_pos in zip(
                anchor_pts[:-1], anchor_pts[1:], rel_poss
            ):
                if rel_pos not in self.should_apply:
                    normal_length += (anchor2 - anchor1).length
                else:
                    vector = self._center - anchor1
                    direction = vector / vector.length
                    delta_t = (prop.length - normal_length) / total_length
                    movement = [] if anchor1 == p else [anchor1 - p]
                    applied_force = (
                        vector
                        if vector.length
                        < self._force
                        * delta_t  # The final movement should not exceed the center.
                        else direction * self._force * delta_t
                    )
                    movement.append(p + prop - anchor1 + applied_force)
                    break
            movement = [prop] if movement is None else movement
            movements.extend(movement)
            for m in movement:
                p += m

        return movements

    def __repr__(self) -> str:
        return (
            f"Black hold zone centered on {self._center} with a force rate {self._force}. "
            + str(self._zone)
        )


class RewardRegion(Region):
    """Reward WILL BE applied when the agent touches the boundary."""

    COLOR = [255, 102, 102]

    def __init__(self, zone: Polygon, reward: float) -> None:
        super().__init__(zone)
        self._reward = reward

    def apply_reward(self, p0: Point, propositions: List[Vector]):
        p = p0
        total_length = 0
        counted_length = 0
        for prop in propositions:
            total_length += prop.length
            anchor_pts, rel_poss = self.check_segment_cross(p, prop)
            for anchor1, anchor2, rel_pos in zip(
                anchor_pts[:-1], anchor_pts[1:], rel_poss
            ):
                if rel_pos in self.should_apply:
                    counted_length += (anchor2 - anchor1).length
            p += prop

        if total_length == 0:
            rel_pos = self._zone.point_relative_pos(p0)
            if rel_pos in self.should_apply:
                return self._reward
            return 0

        return self._reward * counted_length / total_length

    @property
    def should_apply(self):
        return [RelativePos.IN, RelativePos.ON]

    def __repr__(self) -> str:
        return f"Reward zone with reward rate '{self._reward}'. " + str(self._zone)


class SimpleRewardRegion(RewardRegion):
    """Penalizes the agent as long as it touches the region."""

    def apply_reward(self, p0: Point, propositions: List[Vector]):
        rel_pos = self._zone.point_relative_pos(p0 + sum(propositions))
        if rel_pos in self.should_apply:
            return self._reward
        return 0

    def __repr__(self) -> str:
        return f"Simple reward zone with constant reward '{self._reward}/step'. " + str(
            self._zone
        )




from abc import ABC, abstractmethod
from typing import Union, List, Literal
from enum import Enum
import numpy as np


class Element(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def __repr__(self) -> str:
        ...


class Point(Element):
    def __init__(self, x: float, y: float) -> None:
        self._x = x
        self._y = y
        self._pos = np.array([x, y])

    @property
    def pos(self) -> np.array:
        return self._pos

    def __add__(self, vector: "Vector") -> "Point":
        if not isinstance(vector, Vector):
            raise TypeError("Can only add a vector to a point.")
        return Point(*(self._pos + vector._pos))

    def __sub__(self, other: "Point") -> "Vector":
        if not isinstance(other, Point):
            raise TypeError(
                "Can only substract a point from a point to generate a vector."
            )
        return Vector(*(self._pos - other._pos))

    def __eq__(self, other: "Point") -> bool:
        return np.array_equal(self._pos, other._pos)

    def __hash__(self):
        return hash((self._x, self._y))

    def __repr__(self) -> str:
        return "Point at ({x:.2f}, {y:.2f})".format(x=self._x, y=self._y)


class Vector(Point):
    def __init__(self, x: float, y: float) -> None:
        super().__init__(x, y)
        self._length = np.linalg.norm(self._pos)

    @property
    def length(self) -> float:
        return self._length

    def __add__(self, other: "Vector") -> "Vector":
        return Vector(*(self._pos + other._pos))

    def __radd__(self, other) -> "Vector":
        if isinstance(other, int) and other == 0:
            return self
        return self.__add__(other)

    def __sub__(self, other: "Vector") -> "Vector":
        return Vector(*(self._pos - other._pos))

    def __neg__(self):
        return Vector(*(-self._pos))

    def __mul__(self, other: Union[float, "Vector"]) -> "Vector":
        if isinstance(other, Vector):
            return Vector(*(self._pos * other._pos))
        elif isinstance(other, (int, float)):
            return Vector(*(self._pos * other))
        else:
            raise TypeError(f"Unsupported type {type(other)} for __mul__")

    def __rmul__(self, other: Union[float, "Vector"]) -> "Vector":
        return self * other

    def __truediv__(self, a: float) -> "Vector":
        assert isinstance(a, (float, int))
        return Vector(*(self._pos / a))

    def dot(self, other: "Vector") -> float:
        return np.dot(self._pos, other._pos)

    def cross(self, other: "Vector") -> float:
        return np.cross(self._pos, other._pos)

    def __repr__(self) -> str:
        return "Vector ({x:.2f}, {y:.2f})".format(x=self._x, y=self._y)


class Line(Element):
    def __init__(self, a: Point, b: Point) -> None:
        self._a = a
        self._b = b
        self._envelop = (
            Point(*np.minimum(a._pos, b._pos)),
            Point(*np.maximum(a._pos, b._pos)),
        )

    def is_repel(self, other: "Line") -> bool:
        if all(
            (
                self._envelop[0]._x <= other._envelop[1]._x,
                self._envelop[1]._x >= other._envelop[0]._x,
                self._envelop[0]._y <= other._envelop[1]._y,
                self._envelop[1]._y >= other._envelop[0]._y,
            )
        ):
            return False
        return True

    def is_straddle(self, other: "Line") -> bool:
        """
        The criterion is "<=", which implies that the ends of one line can
        be "on" the extension cord of the other. Hence, two lines will be
        considered as "straddle" even if they overlap one another.
        """
        ac = other._a - self._a
        ad = other._b - self._a
        ab = self._b - self._a
        if ac.cross(ab) * ad.cross(ab) <= 0:
            return True
        return False

    def get_overlap(self, other: "Line") -> "Line":
        """
        self: ab   other: cd
        Preserves the direction for self.
        """

        def __check_zero_length(line: "Line"):
            if line._a == line._b:
                return None
            return line

        if self.is_repel(other):
            return None

        ac = other._a - self._a
        ad = other._b - self._a
        ab = self._b - self._a
        if ac.cross(ab) == 0 and ad.cross(ab) == 0:
            c_on = self.point_on(other._a)
            d_on = self.point_on(other._b)
            if c_on and d_on:
                return other
            if (not c_on) and (not d_on):
                return self

            on = other._a if c_on else other._b
            out = other._a if d_on else other._b
            ao = out - self._a
            if ab.dot(ao) > 0:
                return __check_zero_length(Line(on, self._b))
            return __check_zero_length(Line(self._a, on))
        return None

    def is_cross(self, other: "Line") -> bool:
        if self.is_repel(other):
            return False

        if all((self.is_straddle(other), other.is_straddle(self))):
            return True
        return False

    def cross_point(self, other: "Line") -> Union[bool, Point]:
        """
        Finds the cross point if it exists, returns if the two lines
        overlap is the cross point does not exist.

        self: ab   other: cd

        Args:
            other (Line): _description_

        Returns:
            Union[bool, Point]: (Point) if the cross point exists
                                (True) if they overlap
                                (False) if they do not cross over at all
        """
        ca = self._a - other._a
        cb = self._b - other._a
        da = self._a - other._b
        db = self._b - other._b
        S_abc = ca.cross(cb)  # 2 times the area size
        S_abd = da.cross(db)
        S_cda = ca.cross(da)
        S_cdb = cb.cross(db)

        if S_abc == S_abd == 0:
            ab = self._b - self._a
            cd = other._b - other._a
            # On the extension cord
            if self.is_repel(other):
                # Away from each other
                return False
            elif any(
                (
                    self._a == other._a and ab.dot(cd) < 0,
                    self._a == other._b and ab.dot(cd) > 0,
                )
            ):
                return self._a
            elif any(
                (
                    self._b == other._a and ab.dot(cd) > 0,
                    self._b == other._b and ab.dot(cd) < 0,
                )
            ):
                return self._b
            else:
                return True
        elif (S_abc * S_abd > 0) or (S_cda * S_cdb > 0):
            return False
        else:
            t = S_cda / (S_abd - S_abc)
            delta = t * (self._b - self._a)
            return self._a + delta

    def point_on(self, point: Point) -> bool:
        ac = point - self._a
        bc = point - self._b
        if all(
            (
                ac.cross(bc) == 0,
                ac.dot(bc) <= 0,
            )
        ):
            return True
        return False

    def __repr__(self) -> str:
        return "Line ({x1:.2f}, {y1:.2f}), ({x2:.2f}, {y2:.2f})".format(
            x1=self._a._x, y1=self._a._y, x2=self._b._x, y2=self._b._y
        )


class DirectEdge(Line):
    def __init__(self, a: Point, v: Vector) -> None:
        super().__init__(a, a + v)
        self._v = v

    def get_overlap(self, other: Line) -> "DirectEdge":
        """
        This method SHOULD ONLY BE CALLED when you know that they are overlapping!
        """
        raise NotImplementedError


class RelativePos(Enum):
    IN = -1
    ON = 0
    OUT = 1


class Polygon(Element):
    def __init__(self, vertices: List[Point]) -> None:
        assert len(vertices) >= 3
        self._vertices = vertices
        self._edges = [
            Line(a, b) for a, b in zip(vertices, vertices[1:] + [vertices[0]])
        ]
        for e, e_n in zip(self._edges, self._edges[1:] + [self._edges[0]]):
            if (e._b - e._a).cross(e_n._b - e_n._a) == 0:
                raise ValueError(f"Edge {e} and {e_n} are parallel!")
        self._envelop = (
            Point(*np.minimum.reduce([p._pos for p in vertices])),
            Point(*np.maximum.reduce([p._pos for p in vertices])),
        )

    def point_relative_pos(self, point: Point) -> Literal[-1, 0, 1]:
        """
        The logic is ported from (in Chinese):
        https://blog.csdn.net/zsjzliziyang/article/details/108813349
        """
        in_flag = False
        for edge in self._edges:
            a, b = edge._a, edge._b
            if edge.point_on(point):
                return RelativePos.ON

            if any(
                (
                    a._y < point._y and b._y >= point._y,
                    a._y >= point._y and b._y < point._y,
                )
            ):
                # The if above guarantees that a.y != b.y
                x_cross = a._x + (point._y - a._y) * ((b._x - a._x) / (b._y - a._y))
                if x_cross == point._x:
                    return RelativePos.ON
                if x_cross > point._x:
                    in_flag = not in_flag

        return RelativePos.IN if in_flag else RelativePos.OUT

    def __repr__(self) -> str:
        points = [str(p) for p in self._vertices]
        return "Polygon defined by (ordered) vertices: " + ", ".join(points)


class Box(Polygon):
    """
    0  ------------>  x

    |   a -- width -- b
    |
    |   |             |
    |   |           height
    v   |             |

    y   c ----------- d
    """

    def __init__(self, vertex: Point, width: float, height: float) -> None:
        ab = Vector(width, 0)
        ac = Vector(0, height)
        super().__init__([vertex, vertex + ab, vertex + ab + ac, vertex + ac])

    def point_relative_pos(self, point: Point) -> RelativePos:
        if all((
            (point._pos >= self._envelop[0]._pos).all(),
            (point._pos <= self._envelop[1]._pos).all(),
        )):
            for edge in self._edges:
                if edge.point_on(point):
                    return RelativePos.ON
            return RelativePos.IN
        return RelativePos.OUT




import gym
import numpy as np
from gym import spaces
import pygame
from pygame.locals import Rect
from typing import Dict, List

from nav2d.assets.elements import Point, Vector, Line, DirectEdge, RelativePos, Polygon, Box
from nav2d.assets.regions import Region, DynamicRegion, SimpleDynamicRegion, PunchRegion, NoEntryRegion, SlipperyRegion, BlackHoleRegion, RewardRegion, SimpleRewardRegion

from . import EPS_GOAL


class Navigation(gym.Env):
    def __init__(
            self,
            init_x_normal: tuple = (1., 0.5),
            init_y_normal: tuple = (5., 1.25),
            screen_scale: int = 600,
            margin: float = 0.1,
            dynamic_zones: List[DynamicRegion] = [],
            reward_zones: List[RewardRegion] = [],
            draw_init: bool = True,
            dim: int = 2,
            **kwargs,
    ):
        #
        self.v_max = 0.5
        self._draw_init = draw_init
        self._dim = dim

        # The map spans (0, 0) to (10, 10)
        self.low_state = np.array([0, 0] + [0] * (self._dim - 2) + [0, 0], dtype=np.float32)
        self.high_state = np.array([10, 10] + [5] * (self._dim - 2) + [10, 10], dtype=np.float32)
        self.observation_space = spaces.Box(
            low=self.low_state,
            high=self.high_state,
            shape=(self._dim+2, ),
            dtype=np.float32,
        )

        # The largest step size of the agent in one timestep is capped at v_max
        self.min_actions = np.array(
            [-self.v_max] * self._dim, dtype=np.float32,
        )
        self.max_actions = np.array(
            [self.v_max] * self._dim, dtype=np.float32,
        )
        self.action_space = spaces.Box(
            low=self.min_actions,
            high=self.max_actions,
            dtype=np.float32,
        )

        self.goal = np.array([9.0, 9.0])
        self.centers = np.array([5, 5])
        self.init_center = np.array([init_x_normal[0], init_y_normal[0]])

        # Special zones
        self.dynamic_zones = dynamic_zones
        self.reward_zones = reward_zones

        # Initial state distribution
        x_normal, y_normal = np.array(init_x_normal), np.array(init_y_normal)
        self.init_dist = [x_normal, y_normal]

        # Reward setup
        self.time_penalty = 0.5                      # Reward penalty per each timestep
        self.control_penalty = 0.4
        self.goal_dist_penalty = 0.5

        # For rendering
        self.margin = int(screen_scale * margin)
        self.screen_size = [screen_scale + 2 * self.margin, screen_scale + 2 * self.margin]
        self.screen_scale = screen_scale
        self.normalized_state_scale = 1 / (self.high_state[0] - self.low_state[0])
        self.background_color = [255, 255, 255]     # White background
        self.box_color = [128, 128, 128]
        self.init_color = [0, 0, 255]
        self.init_radius = 30                       # The radius of the init region (in pixels)
        self.goal_color = [0, 255, 0]
        self.goal_radius = 15                       # The radius of the goal region (in pixels)
        # self.dynamic_color = [160, 160, 160]        # Set these two in nav_ed.py
        # self.reward_color = [255, 102, 102]
        self.agent_color = [0, 0, 0]                # Black
        self.action_color = [255, 0, 0]
        self.action_mul = 1                         # Action force vector multiplier
        self.movement_color = [0, 0, 0]
        self.movement_mul = 1                       # Movement vector multiplier
        self.agent_radius = 9                       # The radius of the agent represented as a circle (in pixels)
        self.agent_width = 3                        # The width of the line of the circle representing the agent
        self.viewer = None

        self.pos_dim = 2
        self.obs_dim = self.observation_space.low.size
        self.action_dim = self.action_space.low.size

        self.np_random = None

    def seed(self, seed=None):
        self.np_random = np.random.default_rng(seed)
        return [seed]

    def __need_resample(self, pos):
        if not all((
            (self.low_state[:self.pos_dim] <= pos).all(),
            (pos <= self.high_state[:self.pos_dim]).all(),
        )):
            return True
        for zone in self.dynamic_zones:
            if isinstance(zone, NoEntryRegion):
                if zone._zone.point_relative_pos(Point(*pos)) == Polygon.IN:
                    return True
        return False

    def reset(self, eval=False):
        if self.np_random is None:
            self.seed(0)

        # Sample initial position
        sampled = False
        while not sampled:
            # Normal initial state distribution
            self.init_pos = np.array([self.np_random.normal(self.init_dist[0][0], self.init_dist[0][1]),
                                      self.np_random.normal(self.init_dist[1][0], self.init_dist[1][1])], dtype=np.float32)
            if not self.__need_resample(self.init_pos):
                sampled = True
            # self.init_pos = np.clip(self.init_pos, self.low_state, self.high_state)
            # sampled = True
        
        # Add auxiliary dimensions randomly sampled from the standard normal
        if self._dim > 2:
            obs_ = self.np_random.random(size=(self._dim-2, )) \
                    * (self.high_state[self._dim-1] - self.low_state[self._dim-1]) \
                        + self.low_state[self._dim-1]
            self.init_pos = np.concatenate([self.init_pos, obs_])
        
        if eval:
            pass

        # For rendering
        self.last_state = None
        self.action = None

        self.state = np.concatenate([self.init_pos, self.goal])
        return self.state.copy()

    def set_state_from_obs(
            self,
            obs: np.ndarray,
    ):
        self.set_state(obs.copy())

    def get_dist_to_goal(self):
        return np.linalg.norm(self.goal_pos - self.agent_pos)

    def step(self, action: np.ndarray):
        self.last_state = self.state.copy()
        self.action = action.copy()[:self.pos_dim]
        # action = np.clip(action, self.min_actions, self.max_actions)      # Handled by ``NormalizedBoxEnv``
        # assert self.action_space.contains(action)

        # print(f"s: {self.state}, ", end='')
        # Handle the agent's movement
        obs = self.state.copy()
        obs[:self.pos_dim], add_rew = self.transition(self.agent_pos.copy(), 
                                                      self.action.copy())
        
        # Handle dimensions with no semantics
        if self._dim > self.pos_dim:
            obs[self.pos_dim: self._dim] += action.copy()[self.pos_dim: self._dim] + \
                                            self.np_random.normal(0, 0.01, size=(self._dim-self.pos_dim, ))
        
        obs = np.clip(obs, self.low_state, self.high_state)
        self.set_state(obs.copy())

        d_goal = self.get_dist_to_goal()                # R(s', a)
        action_cost = np.linalg.norm(action[:self.pos_dim])     # only the first 2 dimensions are considered
        reward = (
            - self.time_penalty
            - self.control_penalty * action_cost
            - self.goal_dist_penalty * d_goal
            + add_rew
        )

        done = False
        if d_goal < EPS_GOAL:
            done = True
        # print(f"a: {action}, s': {obs}, r: {reward}.")
        return obs.copy(), reward, done, {}

    def transition(self, pos, action: np.array):
        """
        From the current agent position (``pos``) and given an action (``action``),
        computes the next position of the agent. By default, a linear dynamics is assumed.
        But, subject to the current position or the predicted next position,
        non-linear dynamics can also be implemented.
        """
        assert pos.size == action.size
        action = action.squeeze()
        p0 = Point(*pos)
        proposition = [Vector(*action)]
        
        # Applies dynamics according to the distance to the contact point.
        if len(self.dynamic_zones) > 0:
            dist_to_contacts = []
            for zone in self.dynamic_zones:
                dist_to_contacts.append(zone.first_contact(p0, proposition)[1])
            _, self.dynamic_zones = zip(*sorted(zip(dist_to_contacts, self.dynamic_zones), key=lambda x: x[0]))
            for zone in self.dynamic_zones:
                proposition = zone.apply_dynamic(p0, proposition)

        # Applies rewards with an arbitrary order.
        add_rew = 0
        for zone in self.reward_zones:
            add_rew += zone.apply_reward(p0, proposition)

        return (p0 + sum(proposition)).pos, add_rew

    @property
    def agent_pos(self):
        return self.state[:self.pos_dim]

    @property
    def goal_pos(self):
        return self.state[-self.pos_dim:]

    def get_dataset(self) -> Dict:
        """Loads the offline dataset of the related environment: Navigation, NavWithWall, or NavUnsafe."""
        import pickle
        import hydra.utils
        from pathlib import Path
        try:
            cwd = hydra.utils.get_original_cwd()
        except ValueError as e:
            cwd = Path.cwd()
        f_pickled = Path(cwd) / f"data/navigation/{self.config}/{self.config}.pkl"

        with open(f_pickled, 'rb') as file:
            dataset = pickle.load(file)
        return dataset

    def render_for_subclass(self):
        pass

    def render(self, mode='human'):
        if self.viewer is None:
            pygame.init()
            self.viewer = pygame.display.set_mode(self.screen_size)
            self.clock = pygame.time.Clock()

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                exit()

        self.viewer.fill(self.background_color)

        # Draw the state space as a box
        rect = Rect(self.margin, self.margin,
                    self.screen_size[0] - 2 * self.margin,
                    self.screen_size[1] - 2 * self.margin)
        pygame.draw.rect(self.viewer, self.box_color, rect)

        # Draw the initial position
        if self._draw_init:
            init_pos = self.to_pixel(self.init_center)
            pygame.draw.circle(self.viewer, self.init_color, init_pos, self.init_radius)

        # Draw the goal position
        goal_pos = self.to_pixel(self.goal)
        pygame.draw.circle(self.viewer, self.goal_color, goal_pos, self.goal_radius)

        # Draw dynamic regions
        for region in self.dynamic_zones:
            region.render(self.viewer, self.to_pixel)

        # Draw reward regions
        for region in self.reward_zones:
            region.render(self.viewer, self.to_pixel)

        # Draw any other registered objects on screen
        self.render_for_subclass()

        # Draw the agent position
        pos = self.to_pixel(self.agent_pos)
        pygame.draw.circle(self.viewer, self.agent_color, pos, self.agent_radius, self.agent_width)

        # Try to draw the previous action
        if self.action is not None:
            start = Point(*self.last_state[:self.pos_dim])
            end = start + Vector(*self.action) * self.action_mul
            pygame.draw.line(
                self.viewer,
                self.action_color,
                self.to_pixel(start.pos),
                self.to_pixel(end.pos)
            )

        # Try to draw the movement
        if self.last_state is not None:
            start = Point(*self.last_state[:self.pos_dim])
            end = start + (Point(*self.state[:self.pos_dim]) - start) * self.movement_mul
            pygame.draw.line(
                self.viewer,
                self.movement_color,
                self.to_pixel(start.pos),
                self.to_pixel(end.pos)
            )

        # Update the screen
        pygame.display.update()

        if mode == 'rgb_array':
            data = pygame.surfarray.array3d(self.viewer)
            data = np.transpose(data, (1, 0, 2))
            return data
        # plt.imshow(np.transpose(arr, (1, 0, 2)), interpolation='nearest', origin='lower')
        # plt.show()

        # Time delay
        self.clock.tick(10)

    def to_pixel(self, x):
        return (self.screen_scale * x * self.normalized_state_scale + self.margin).astype(int).tolist()

    def set_state(self, state):
        self.state[:] = state

    def get_goal_from_obs(self, obs):
        """obs = (x, y, ..., gx, gy)"""
        return obs[..., -2:]




from setuptools import setup
from setuptools import find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='nav2d',
    packages=find_packages(),
    version='0.0.1',
    license='MIT License',
    description='2D Navigation Gym environment for Reinforcement Learning',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Xiaoyu Wang',
    author_email='cnxiaoyu.wang@mail.utoronto.ca',
    url='https://github.com/xxArbiter/nav2d',
    download_url="https://github.com/xxArbiter/nav2d/archive/refs/tags/0.0.2.tar.gz"
)